use std::str::FromStr;
use std::time::Duration;

use log::{error, info};
use rumqttc::{Event, Incoming, MqttOptions, QoS};
use tokio::sync::{broadcast, mpsc};

use deckster_shared::handler_communication::{HandlerCommand, HandlerEvent};
use deckster_shared::path::{KeyPath, KeyPosition, KnobPath, KnobPosition};

use crate::model::mqtt::HandlerHostsConfig;
use crate::model::mqtt::MqttConfig;

pub async fn start_mqtt_client(
    config: &MqttConfig,
    handler_hosts_config_sender: mpsc::Sender<Option<HandlerHostsConfig>>,
    commands_receiver: flume::Receiver<HandlerCommand>,
    events_sender: broadcast::Sender<HandlerEvent>,
) {
    let topic_prefix = config.topic_prefix.to_owned();

    let mut options = MqttOptions::new(&config.client_id, &config.host, config.port);
    options.set_keep_alive(Duration::from_secs(3));
    options.set_clean_session(true);

    if let Some(credentials) = &config.credentials {
        options.set_credentials(&credentials.username, &credentials.password);
    }

    let (client, mut event_loop) = rumqttc::AsyncClient::new(options, 32);

    tokio::spawn({
        let topic_prefix = topic_prefix.clone();
        let client = client.clone();

        async move {
            let mut is_connected = false;
            let mut is_first_try = true;

            loop {
                match event_loop.poll().await {
                    Err(error) => {
                        if is_first_try {
                            error!("MQTT connection failed (retrying in the background): {error}");
                            is_first_try = false;
                        } else if is_connected {
                            error!("MQTT connection lost: {error}");
                            handler_hosts_config_sender.send(None).await.unwrap();
                        }

                        is_connected = false;
                        tokio::time::sleep(Duration::from_secs(2)).await;
                    }
                    Ok(Event::Incoming(event)) => {
                        match event {
                            Incoming::ConnAck(_) => {
                                info!("MQTT connection established.");
                                is_connected = true;
                                is_first_try = false;

                                client.subscribe(format!("{topic_prefix}/config"), QoS::ExactlyOnce).await.unwrap();
                                client.subscribe(format!("{topic_prefix}/keys/+/+/events"), QoS::ExactlyOnce).await.unwrap();
                                client.subscribe(format!("{topic_prefix}/knobs/+/+/events"), QoS::ExactlyOnce).await.unwrap();
                            }
                            Incoming::Publish(event) => {
                                let topic_segments = event.topic.strip_prefix(&topic_prefix).unwrap().split('/').skip(1).collect::<Vec<&str>>();

                                if topic_segments[0] == "config" {
                                    if let Ok(config) = serde_json::from_slice(&event.payload) {
                                        handler_hosts_config_sender.send(config).await.unwrap();
                                    } else {
                                        log::error!("Could not deserialize the latest configuration from {}", event.topic);
                                        handler_hosts_config_sender.send(None).await.unwrap();
                                    };
                                } else {
                                    let page_id = topic_segments[1];
                                    let position = topic_segments[2];
                                    let property = topic_segments[3];

                                    match topic_segments[0] {
                                        "keys" if property == "events" => {
                                            if let Ok(event) = serde_json::from_slice(&event.payload) {
                                                events_sender
                                                    .send(HandlerEvent::Key {
                                                        path: KeyPath {
                                                            page_id: page_id.to_owned(),
                                                            position: KeyPosition::from_str(position).unwrap(),
                                                        },
                                                        event,
                                                    })
                                                    .ok(); // This can be Err when events are received before the configuration
                                            } else {
                                                log::error!("Could not deserialize the latest event from {}", event.topic);
                                            };
                                        }
                                        "knobs" if property == "events" => {
                                            if let Ok(event) = serde_json::from_slice(&event.payload) {
                                                events_sender
                                                    .send(HandlerEvent::Knob {
                                                        path: KnobPath {
                                                            page_id: page_id.to_owned(),
                                                            position: KnobPosition::from_str(position).unwrap(),
                                                        },
                                                        event,
                                                    })
                                                    .ok(); // This can be Err when events are received before the configuration
                                            } else {
                                                log::error!("Could not deserialize the latest event from {}", event.topic);
                                            };
                                        }
                                        _ => {}
                                    }
                                }
                            }
                            _ => {}
                        }
                    }
                    _ => {}
                }
            }
        }
    });

    tokio::spawn(async move {
        while let Ok(command) = commands_receiver.recv_async().await {
            match command {
                HandlerCommand::SetActivePages { .. } => log::warn!("HandlerCommand::SetActivePages is not supported for remote handlers."),
                HandlerCommand::SetKeyStyle { path, value } => client
                    .publish(
                        format!("{topic_prefix}/keys/{path}/style"),
                        QoS::ExactlyOnce,
                        false,
                        serde_json::to_vec(&value).unwrap(),
                    )
                    .await
                    .unwrap(),
                HandlerCommand::SetKnobStyle { path, value } => client
                    .publish(
                        format!("{topic_prefix}/knobs/{path}/style"),
                        QoS::ExactlyOnce,
                        false,
                        serde_json::to_vec(&value).unwrap(),
                    )
                    .await
                    .unwrap(),
                HandlerCommand::SetKnobValue { path, value } => client
                    .publish(
                        format!("{topic_prefix}/knobs/{path}/value"),
                        QoS::ExactlyOnce,
                        false,
                        serde_json::to_vec(&value).unwrap(),
                    )
                    .await
                    .unwrap(),
            };
        }
    });
}
